import java.io.*;

/**
* A class to represent the CRSPlus A150 robot arm system
* @author Ho Ng
* @version 01/09/1998
*/
public class CRSPlusA150 extends DHSystem
{
    private static double MAXERROR    = 1e-10;

    private static double[] MAXJOINT = {degToRad(175.0) + MAXERROR,
                                        degToRad(110.0) + MAXERROR,
                                        MAXERROR,
                                        degToRad(110.0) + MAXERROR,
                                        degToRad(180.0) + MAXERROR};
    private static double[] MINJOINT = {degToRad(-175.0) - MAXERROR,
                                        - MAXERROR,
                                        degToRad(-125.0) - MAXERROR,
                                        degToRad(-110.0) - MAXERROR,
                                        degToRad(-180.0) - MAXERROR};
    private static double MAXGRIP    = 2.0 + MAXERROR;
    private static double MINGRIP    = 0.0 - MAXERROR;
    private static int MAXSPEED      = 50;
    private static int MINSPEED      = 0;

    /**
    * purppose : convert angle from degree into radian
    * @param degree angle in degree
    * @return angle in radian
    */
    protected static double degToRad(double degree)
    {
        return (degree / 180.0 * Math.PI);
    }

    /**
    * purpose : convert angle from radian into degree
    * @param radian angle in radian
    * @return angle in degree
    */
    protected static double radToDeg(double radian)
    {
        return (radian * 180.0 / Math.PI);
    }

    /**
    * purpose : change all joint angles
    * @param waist waist angle
    * @param shoulder shoulder angle
    * @param elbow elbow angle
    * @param wrist wrist bend angle
    * @param twist wrist twist angle
    * @return true (succeed) or false (fail)
    */
    private boolean changeJoints(double waist, double shoulder, double elbow,
                                 double wrist, double twist)
    {
        if ((waist > MAXJOINT[0]) || (waist < MINJOINT[0]))
        {
            System.out.println("CRS : waist angle is out of the range");
            return false;
        }

        if ((shoulder > MAXJOINT[1]) || (shoulder < MINJOINT[1]))
        {
            System.out.println("CRS : shoulder angle is out of the range");
            return false;
        }

        if ((elbow > MAXJOINT[2]) || (elbow < MINJOINT[2]))
        {
            System.out.println("CRS : elbow angle is out of the range");
            return false;
        }

        if ((wrist > MAXJOINT[3]) || (wrist < MINJOINT[3]))
        {
            System.out.println("CRS : wrist angle is out of the range");
            return false;
        }

        if ((twist > MAXJOINT[4]) || (twist < MINJOINT[4]))
        {
            System.out.println("CRS : twist angle is out of the range");
            return false;
        }

        changeTheta(1, waist);
        changeTheta(2, shoulder);
        changeTheta(3, elbow);
        changeTheta(4, wrist);
        changeTheta(5, twist);
        return true;
    }

    /**
    * purpose : change the width between the fingers of the gripper
    * @param width width between the fingers
    * @return true (succeed) or false (fail)
    */
    private boolean changeFingers(double width)
    {
        if ((width < MINGRIP) || (width > MAXGRIP))
        {
            System.out.println("CRS : fingers of the gripper are out of " +
                               "the range");
            return false;
        }

        double distance = width / 2.0 ;
        changeD(6, distance);
        changeD(7, distance);
        return true;
    }

    /**
    * Default constructor
    * purpose : initialize the robot arm system includes the gripper
    */
    public CRSPlusA150()
    {
        super(7);

        /* initialize the coordinate system of the arm */
        double angle = 0.5 * Math.PI;
        setDH(1, 0.0, 0.0, 0.0, 10.0, 0);
        setDH(2, angle, 0.0, 0.0, 0.0, 1);
        setDH(3, 0.0, 10.0, 0.0, 0.0, 2);
        setDH(4, 0.0, 10.0, angle, 0.0, 3);
        setDH(5, angle, 0.0, 0.0, 5.5, 4);
        setDH(6, -angle, 0.0, 0.0, 0.0, 5);
        setDH(7, angle, 0.0, 0.0, 0.0, 5);
    }

    /**
    * purpose : home the robot arm
    */
    public void home()
    {
        // no function
    }

    /**
    * purpose : set the arm to ready position
    */
    public void ready()
    {
        double angle = 0.5 * Math.PI;
        changeJoints(0.0, angle, -angle, 0.0, 0.0);
        changeFingers(2.0);
    }

    /**
    * purpose : wait for the previous motion instruction to complete
    */
    public void finish()
    {
       // no-op
    }
    
    /**
    * purpose : execute the command MA of the arm
    * @param waist waist angle in radian
    * @param shoulder shoulder angle in radian
    * @param elbow elbow angle in radian
    * @param wrist wrist's bend angle in radian
    * @param twist wrist's twist angle in radian
    * @return true (succeed) or false (fail)
    */
    public boolean ma(double waist, double shoulder, double elbow,
                      double wrist, double twist)
    {
        return changeJoints(waist, shoulder, elbow, wrist, twist);
    }

    /**
    * purpose : execute the command MA of the arm
    * @param waist waist angle in degree
    * @param shoulder shoulder angle in degree
    * @param elbow elbow angle in degree
    * @param wrist wrist's bend angle in degree
    * @param twist wrist's twist angle in degree
    * @return true (succeed) or false (fail)
    */
    public boolean madeg(double waist, double shoulder, double elbow,
                         double wrist, double twist)
    {
        return changeJoints(degToRad(waist), degToRad(shoulder),
                            degToRad(elbow), degToRad(wrist), degToRad(twist));
    }

    /**
    * purpose : execute the command MI of the arm
    * @param waist increment of waist angle in radian
    * @param shoulder increment of shoulder angle in radian
    * @param elbow increment of elbow angle in radian
    * @param wrist increment of wrist's bend angle in radian
    * @param twist increment of wrist's twist angle in radian
    * @return true (succeed) or false (fail)
    */
    public boolean mi(double waist, double shoulder, double elbow,
                      double wrist, double twist)
    {
        return changeJoints(getTheta(1) + waist, getTheta(2) + shoulder,
                            getTheta(3) + elbow, getTheta(4) + wrist,
                            getTheta(5) + twist);
    }

    /**
    * purpose : execute the command MI of the arm
    * @param waist increment of waist angle in degree
    * @param shoulder increment of shoulder angle in degree
    * @param elbow increment of elbow angle in degree
    * @param wrist increment of wrist's bend angle in degree
    * @param twist increment of wrist's twist angle in degree
    * @return true (succeed) or false (fail)
    */
    public boolean mideg(double waist, double shoulder, double elbow,
                         double wrist, double twist)
    {
        return mi(degToRad(waist),
                  degToRad(shoulder),
                  degToRad(elbow),
                  degToRad(wrist),
                  degToRad(twist));
    }

    /**
    * to be implemented
    */
    public void move(double x0, double y0, double z0)
    {
        // to be implemented
    }

    /**
    * purpose : execute the command GRIP of the arm
    * @param width width between the fingers of the arm in inch
    * @return true (succeed) or false (fail)
    */
    public boolean grip(double width)
    {
        return changeFingers(width);
    }

    /**
    * purpose : execute the command SPEED of the arm
    * @param value speed of the arm
    * @return true (succeed) or false (fail)
    */
    public boolean speed(int value)
    {
        if ((value < MINSPEED) || (value > MAXSPEED))
        {
            System.out.println("CRS.speed : speed out of the range");
            return false;
        }

        return true;
    }
};
