import java.awt.event.*;
import java.awt.*;

/**
* A class to display 3D objects on canvas
* @author Ho Ng
* @version 01/09/1998
*/
public class Pic3DCanvas extends Canvas
{
    private static final int DEFAULTUSIZE = 100;
    private static final int DEFAULTVSIZE = 100;
    private static final int MAXOBJS      = 10;

    private Object3D[] obj = new Object3D[MAXOBJS];
    private Matrix[] ctm   = new Matrix[MAXOBJS];

    private Scene pen;
    private Image buffer;
    private int usize, vsize;    /* actual size of the canvas */
    private int total;           /* total number of objects */

    /**
    * Default constructor
    * purpose : initialize the object and set the output canvas's size to 100x100
    */
    public Pic3DCanvas()
    {
        this(DEFAULTUSIZE, DEFAULTVSIZE);
    }

    /**
    * Constructor
    * purpose : initialize the object
    * @param u horizontal size of the output area
    * @param v vertical size of the output area
    */
    public Pic3DCanvas(int u, int v)
    {
        total = 0;

        /* create the output window */
        usize = ((u < DEFAULTUSIZE) ? DEFAULTUSIZE : u);
        vsize = ((v < DEFAULTVSIZE) ? DEFAULTVSIZE : v);
        setSize(usize, vsize);

        /* set the camera */
        pen = new Scene(usize, vsize);
        pen.setCamera(new Vector3D(1.0, 0.0, 0.0), new Vector3D(0.0, 0.0, 0.0),
                      new Vector3D(0.0, 0.0, 1.0));
    }

    /**
    * purpose : add a new Object3D into the scene
    * @param x Object3D
    * @return index of the input Object3D
    */
    public int addObject3D(Object3D x)
    {
        if (total == MAXOBJS)
            return (-1);

        int index = total;
        obj[index] = new Object3D(x);
        ctm[index] = Matrix.identity();
        total++;
        return index;
    }

    /**
    * purpose : change the current transformational matrix of an object
    * @param index index of the object
    * @param m new current transformational matrix
    * @return true (succeed) or false (fail)
    */
    public boolean changeCTM(int index, Matrix m)
    {
        if ((index < 0) || (index >= total))
            return false;

        ctm[index] = new Matrix(m);
        return true;
    }

    /**
    * purpose : change the position and orientation of the camera
    * @param lookFrom position of the camera
    * @param lookAt target's position of the camera
    * @param lookUp orientation of the camera
    * @return true (succeed) or false (fail)
    */
    public boolean updateCamera(Vector3D lookFrom, Vector3D lookAt, Vector3D lookUp)
    {
        return (pen.setCamera(lookFrom, lookAt, lookUp));
    }

    /**
    * purpose : show the scene on the output canvas
    */
    public void display()
    {
        if (buffer == null)
        {
            buffer = createImage(usize, vsize);
            if (buffer == null)
                return;
        }

        /* draw the output picture */
        Graphics g = buffer.getGraphics();
        g.setColor(Color.blue);
        g.fillRect(0, 0, usize, vsize);
	g.setColor(Color.white);

        for (int i = 0; i < total; i++)
        {
            pen.setCTM(ctm[i]);
            pen.drawObject3D(obj[i], g);
        }

        repaint();
    }

    /**
    * purpose : get the total number of objects
    * @return total number of objects
    */
    public int numberOfObjects()
    {
        return total;
    }

    /**
    * purpose : update the canvas
    */
    public void update(Graphics g)
    {
        paint(g);
    }

    /**
    * purpose : draw the buffer image on the canvas
    * @param g Graphics of the output canvas
    */
    public void paint(Graphics g)
    {
//	System.out.println("Repainting 3d canvas");
//        g.clearRect(0, 0, usize, vsize);
        if (buffer != null)
            g.drawImage(buffer, 0, 0, this);
//	System.out.println("Repainting 3d canvas (done)");
    }
}
