
import java.awt.*;
import java.io.*;

/**
* A class to provide methods to draw objects in 3D world onto a Graphics object <br>
* version : <br>
* 01/10/1997 original program in C written by Eleanor Yu and Ho Ng <br>
* 09/08/1998 translated into Java by Ho Ng <br>
* 12/08/1998 modified by Ho Ng <br>
* 01/09/1998 modified by Ho Ng to support classes Polygon3D and Object3D
*/
public class Scene
{
    private static final Vector3D DEFAULTLOOKFROM = new Vector3D(0.0, -100.0, 0.0);
    private static final Vector3D DEFAULTLOOKAT   = new Vector3D(0.0, 0.0, 0.0);
    private static final Vector3D DEFAULTLOOKUP   = new Vector3D(0.0, 0.0, 1.0);
    private static Matrix perCVVToVPlane = new Matrix();

    private Matrix ctm;
    private Matrix worldToEye;
    private Matrix eyeToPerCVV;
    private Matrix modelToPerCVV;
    private Matrix vPlaneToVport;
    private Matrix ndcToScreen;
    private Matrix vPlaneToScreen;

    private double zmin;

    /**
    * purpose : update the model-to-perspective transformational matrix
    */
    private void updateModelToPerCVV()
    {
        modelToPerCVV = Matrix.multi(eyeToPerCVV, Matrix.multi(worldToEye, ctm));
    }

    /**
    * purpose : update the perspective projection matrix
    */
    private Polygon3D perspectiveProjection(Polygon3D p)
    {
        return (Polygon3D.transformPolygon(perCVVToVPlane, p));
    }

    /**
    * Default constructor
    * purpose : initialize the Scene object and set the output screen size to 512x512
    */
    public Scene()
    {
        this(512,512);
    }

    /**
    * Constructor
    * purpose : initialize the Scene object
    * @param usize total number of pixels horizontally
    * @param vsize total number of pixels vertically
    */
    public Scene(int usize, int vsize)
    {
        perCVVToVPlane.m[0][0] = 1.0;
        perCVVToVPlane.m[1][1] = 1.0;
        perCVVToVPlane.m[3][2] = 1.0;

        ctm           = Matrix.identity();
        worldToEye    = Matrix.identity();
        eyeToPerCVV   = Matrix.identity();
        vPlaneToVport = Matrix.identity();
        ndcToScreen   = Matrix.identity();

        setCamera(DEFAULTLOOKFROM, DEFAULTLOOKAT, DEFAULTLOOKUP);
        setViewVolume(1.0, 1.0, 1.0, 0.1, 1000.0);
        setViewport(0.0, 0.0, 1.0, 1.0);
        setScreen(usize, vsize);
    }

    /**
    * purpose : change the current transformational matrix
    * @param m new current transformational matrix
    */
    public void setCTM(Matrix m)
    {
        ctm = new Matrix(m);
        updateModelToPerCVV();
    }   

    /**
    * purpose : set the position and orientation of the carmera
    * @param lookFrom position of the camera
    * @param lookAt target's position of the camera
    * @param lookUp orientation of the camera
    */
    public boolean setCamera(Vector3D lookFrom, Vector3D lookAt, Vector3D lookUp)
    {
        Vector3D n = Vector3D.minus(lookAt, lookFrom);
        if (n.normalize() == 0.0)    /* lookFrom == lookAt */
            return false;

        Vector3D u = Vector3D.cross(n, lookUp);
        if (u.normalize() == 0.0)    /* lookUp == (0,0,0) || n == c*lookUp */
            return false;

        Vector3D v = Vector3D.cross(u, n);

        worldToEye.m[0][0] = u.v[0];
        worldToEye.m[0][1] = u.v[1];
        worldToEye.m[0][2] = u.v[2];
        worldToEye.m[0][3] = -(Vector3D.dot(u, lookFrom));
        worldToEye.m[1][0] = v.v[0];
        worldToEye.m[1][1] = v.v[1];
        worldToEye.m[1][2] = v.v[2];
        worldToEye.m[1][3] = -(Vector3D.dot(v, lookFrom));
        worldToEye.m[2][0] = n.v[0];
        worldToEye.m[2][1] = n.v[1];
        worldToEye.m[2][2] = n.v[2];
        worldToEye.m[2][3] = -(Vector3D.dot(n, lookFrom));

        updateModelToPerCVV();
        return true;
    }

    /**
    * purpose : set the view volume of the camera
    * @param width width
    * @param height height
    * @param distance distance
    * @param near near
    * @param far far
    */
    public void setViewVolume(double width, double height, double distance,
                              double near, double far)
    {
        eyeToPerCVV.m[0][0] = 2.0 * distance / (far * width);
        eyeToPerCVV.m[1][1] = 2.0 * distance / (far * height);
        eyeToPerCVV.m[2][2] = 1.0 / far;

        zmin = near / far;

        updateModelToPerCVV();
    }

    /**
    * purpose : set the viewport of the output
    */
    public void setViewport(double llx, double lly, double urx, double ury)
    {
        vPlaneToVport = Matrix.multi(Matrix.translate(1.0, 1.0, 0.0),
                                     Matrix.identity());
        vPlaneToVport = Matrix.multi(Matrix.scale((urx-llx)/2.0, (ury-lly)/2.0,
                                     1.0), vPlaneToVport);
        vPlaneToVport = Matrix.multi(Matrix.translate(llx, lly, 0.0),
                                     vPlaneToVport);

        vPlaneToScreen = Matrix.multi(ndcToScreen, vPlaneToVport);
    }

    /**
    * purpose : set the output screen
    * @param maxU total number of pixels of the screen horizontally
    * @param maxV total number of pixels of the screen vertically
    */
    public void setScreen(int maxU, int maxV)
    {
        ndcToScreen = Matrix.identity();
        ndcToScreen.m[0][0] = (double) maxU;
        ndcToScreen.m[1][1] = (double) (-maxV);
        ndcToScreen.m[1][3] = (double) maxV;

        vPlaneToScreen = Matrix.multi(ndcToScreen, vPlaneToVport);
    }

    /**
    * purpose : project and draw the Polygon3D onto the Graphics
    * @param poly Polygon3D
    * @param g Graphics
    */
    public void drawPolygon3D(Polygon3D poly, Graphics g)
    {
        Polygon3D temp;
        int i;

        temp = Polygon3D.transformPolygon(modelToPerCVV, poly);
        temp = perspectiveProjection(temp);
        temp = Polygon3D.transformPolygon(vPlaneToScreen, temp);

        for (i = 0; i < temp.count - 1; i++)
            g.drawLine((int)(temp.c[i].v[0]), (int)(temp.c[i].v[1]),
                       (int)(temp.c[i+1].v[0]), (int)(temp.c[i+1].v[1]));
        if (i > 0)
            g.drawLine((int)(temp.c[i].v[0]), (int)(temp.c[i].v[1]),
                       (int)(temp.c[0].v[0]), (int)(temp.c[0].v[1]));
    }

    /**
    * purpose : project and draw the Object3D onto the Graphics
    * @param obj Object3D
    * @param g Graphics
    */
    public void drawObject3D(Object3D obj, Graphics g)
    {
        for (int i = 0; i < obj.count; i++)
            drawPolygon3D(obj.p[i], g);
    }
};
